import bisect
import math
import typing
import shapely as sp
from .util import polygons_to_sides, theta_in_between
from .geometry import distance2, move, atan2


class Visibility:
    def __init__(self, arena: sp.Polygon, occlusions: typing.List[sp.Polygon]):
        self.walls, self.walls_centroids, self.vertices, self.walls_vertices = polygons_to_sides(occlusions + [arena])
        arena_vertices = [sp.Point(c) for c in arena.exterior.coords[1:]]
        self.max_distance = max([v1.distance(v2) for v1 in arena_vertices for v2 in arena_vertices])

    def walls_by_distance(self,
                          src: sp.Point):
        return sorted([(side_number,
                        vertices,
                        distance2(src, self.walls_centroids[side_number]))
                       for side_number, vertices
                       in enumerate(self.walls_vertices)],
                      key=lambda item: item[2])

    def intersection(self,
                     src: sp.Point,
                     theta: float,
                     walls_by_distance=None):

        if walls_by_distance is None:
            walls_by_distance = self.walls_by_distance(src=src)

        ray_end = move(src, theta, self.max_distance)
        ray = sp.LineString([src, ray_end])

        for side_number, vertices, distance in walls_by_distance:
            side = self.walls[side_number]
            if ray.intersects(side):
                return ray.intersection(side)
        return None

    def line_of_sight(self,
                      src: typing.Union[sp.Point,typing.Tuple[float, float]],
                      dst: typing.Union[sp.Point,typing.Tuple[float, float]],
                      walls_by_distance=None) -> bool:
        if isinstance(src, tuple):
            src = sp.Point(*src)

        if isinstance(dst, tuple):
            dst = sp.Point(*dst)

        if walls_by_distance is None:
            walls_by_distance = self.walls_by_distance(src=src)

        target_distance = distance2(src,dst)
        ray = sp.LineString([src, dst])
        for side_number, vertices, distance in walls_by_distance:
            if distance >= target_distance:
                return True
            if ray.intersects(self.walls[side_number]):
                return False
        return True

    def get_visibility_polygon(self,
                               location: typing.Tuple[float, float],
                               direction: float,
                               view_field: float = 360):

        start = sp.Point(location)

        if view_field >= 360:
            view_field_radians = None
            direction_radians = None
            view_field_start = None
            view_field_end = None
        else:
            view_field_radians = math.radians(view_field)
            direction_radians = math.radians(direction)
            view_field_start = direction_radians - view_field / 2
            if view_field_start < -math.pi:
                view_field_start += math.pi * 2
            view_field_end = direction_radians + view_field / 2
            if view_field_end > math.pi:
                view_field_end += math.pi * 2

        def theta_in_view_field(theta):
            if view_field_radians is None:
                return True
            return theta_in_between(theta, view_field_start, view_field_end)

        walls_by_distance = self.walls_by_distance(src=start)

        # contains the relative angle in radians between the start point and the occlusion_vertices
        vertices_theta: typing.List[float] = []
        # contains the relative distance between the start point and the occlusion_vertices
        vertices_distances: typing.List[typing.Optional[float]] = []
        # contains the cached visibility information, if None, it must be computed
        vertices_visible: typing.List[typing.Optional[bool]] = []
        for vertex in self.vertices:
            vertex_theta = atan2(start, vertex)
            vertices_theta.append(vertex_theta)
            if theta_in_view_field(vertex_theta):
                vertices_distances.append(distance2(start, vertex))
                vertices_visible.append(None)
            else:
                vertices_distances.append(None)
                vertices_visible.append(False)

        # creates an index of vertices sorted by relative angle to accelerate search
        vertices_by_thetas = sorted([(vertex_number, theta)
                                     for vertex_number, theta
                                     in enumerate(vertices_theta)],
                                    key=lambda item: item[1])

        #  used to track zones in the view field that are
        # covered by occlusion sides. this speeds up the rendering as everything behind
        # them would be tagged as not visible
        def set_vertices_to_not_visible(theta_start, theta_end, distance):
            start_index = bisect.bisect_left(vertices_by_thetas,
                                             theta_start,
                                             key=lambda x: x[1])
            end_index = bisect.bisect_left(vertices_by_thetas,
                                           theta_end,
                                           key=lambda x: x[1])
            if theta_start < theta_end:
                for vertex_number, theta in vertices_by_thetas[start_index: end_index]:
                    if (vertices_visible[vertex_number] is None and
                       vertices_distances[vertex_number] > distance and
                       theta_in_between(theta, theta_start, theta_end)):
                        vertices_visible[vertex_number] = False
            else:
                for vertex_number, theta in vertices_by_thetas[start_index:]:
                    if (vertices_visible[vertex_number] is None and
                       vertices_distances[vertex_number] > distance and
                       theta_in_between(theta, theta_start, theta_end)):
                        vertices_visible[vertex_number] = False
                for vertex_number, theta in vertices_by_thetas[:end_index]:
                    if (vertices_visible[vertex_number] is None and
                       vertices_distances[vertex_number] > distance and
                       theta_in_between(theta, theta_start, theta_end)):
                        vertices_visible[vertex_number] = False

        def solve_thetas(vertices):
            index1, index2 = vertices
            theta1 = vertices_theta[index1]
            theta2 = vertices_theta[index2]
            if theta1 > theta2:
                theta1, theta2 = theta2, theta1
            d1 = theta2 - theta1
            if d1 > math.pi:
                return (index2, theta2), (index1, theta1)
            else:
                return (index1, theta1), (index2, theta2)

        def is_vertex_visible(src: sp.Point,
                              vertex_number,
                              sorted_sides) -> bool:
            dst = self.vertices[vertex_number]
            target_distance = distance2(src, dst)
            ray = sp.LineString([src, dst])
            for side_number, vertices, distance in sorted_sides:
                if vertex_number in vertices:
                    continue
                if distance >= target_distance:
                    return True
                if ray.intersects(self.walls[side_number]):
                    return False
            return True

        connections = []
        for _, vertices, distance in walls_by_distance:
            (index_a, theta_a), (index_b, theta_b) = solve_thetas(vertices)

            if vertices_visible[index_a] is None:
                vertices_visible[index_a] = is_vertex_visible(start, index_a, walls_by_distance)
            a_visible = vertices_visible[index_a]

            if vertices_visible[index_b] is None:
                vertices_visible[index_b] = is_vertex_visible(start, index_b, walls_by_distance)
            b_visible = vertices_visible[index_b]

            if a_visible and b_visible:
                # the whole side of the occlusion is visible, I'll cover that theta range for future use
                set_vertices_to_not_visible(theta_a, theta_b, distance)
                connections.append((index_a, index_b))

        final_vertices = []

        def is_connected(a, b):
            for ca, cb in connections:
                if ca == a and cb == b:
                    return True
                if ca == b and cb == a:
                    return True
            return False

        last_vertex_number, last_theta = vertices_by_thetas[-1]
        for vertex_number, theta in vertices_by_thetas:
            if vertices_visible[vertex_number]:
                vertex = self.vertices[vertex_number]
                if not is_connected(last_vertex_number, vertex_number):
                    ext1 = self.intersection(start, last_theta + 0.001, walls_by_distance)
                    final_vertices.append(ext1)
                    ext2 = self.intersection(start, theta - 0.001, walls_by_distance)
                    final_vertices.append(ext2)
                final_vertices.append(vertex)
                last_vertex_number, last_theta = vertex_number, theta
        return sp.Polygon(final_vertices), final_vertices




import math
import shapely as sp


def distance2(point1: sp.Point, point2: sp.Point):
    return (point2.x - point1.x) ** 2 + (point2.y - point1.y) ** 2


def atan2(src: sp.Point,
          dst: sp.Point):
    return math.atan2(dst.y - src.y, dst.x - src.x)


def move(src: sp.Point,
         theta: float,
         dist: float):
    return sp.Point(src.x + math.cos(theta) * dist, src.y + math.sin(theta) * dist)



import os


class Resources(object):
    path = os.path.dirname(os.path.abspath(__file__))

    @staticmethod
    def file(file_name: str) -> str:
        return os.path.join(Resources.path, "files", file_name)



import shapely as sp
import time
import random
import typing
import pygame
from .agent import AgentState
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources


class Robot(NavigationAgent):
    def __init__(self,
                 start_locations: typing.List[typing.Tuple[float, float]],
                 open_locations: typing.List[typing.Tuple[float, float]],
                 navigation: Navigation):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=0.075,
                                 max_turning_speed=3.5)
        self.start_locations = start_locations
        self.open_locations = open_locations
        self.last_destination_time = 0

    def reset(self):
        NavigationAgent.reset(self)
        self.set_state(AgentState(location=random.choice(self.start_locations),
                                  direction=180))

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("predator.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_polygon() -> sp.Polygon:
        return sp.Polygon([(.02, 0.013), (-.02, 0.013), (-.02, -0.013), (.02, -0.013), (.025, -0.01), (.025, 0.01)])

    def step(self, delta_t: float):
        NavigationAgent.step(self=self,
                             delta_t=delta_t)



import typing
import pygame
from .agent import AgentState
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources
import shapely as sp


class Mouse(NavigationAgent):
    def __init__(self,
                 start_state: AgentState,
                 navigation: Navigation):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=0.5,
                                 max_turning_speed=20.0)
        self.start_state = start_state

    def reset(self):
        NavigationAgent.reset(self)
        self.set_state(AgentState(location=self.start_state.location,
                                  direction=self.start_state.direction))

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("prey.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_polygon() -> sp.Polygon:
        return sp.Polygon([(.015, 0), (0, 0.005), (-.015, 0), (0, -0.005)])




import math
import typing
import shapely as sp
import colorsys


def distance2(point1: typing.Tuple[float, float],
              point2: typing.Tuple[float, float]):
    return (point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2


def generate_distinct_colors(n):
    """
    Generate a list of n distinct RGB colors.

    Parameters:
    n (int): The number of distinct colors to generate.

    Returns:
    List[Tuple[int, int, int]]: A list of n RGB colors.
    """
    colors = []
    for i in range(n):
        hue = i / n  # Evenly distribute hues in the [0, 1) interval
        saturation = 0.7  # Choose a saturation level that avoids white and grays
        lightness = 0.5  # Choose a lightness level that avoids black and white
        rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
        # Convert the RGB values from [0, 1] to [0, 255] and round them to integers
        rgb = tuple(round(c * 255) for c in rgb)
        colors.append(rgb)

    return colors


def create_hexagon(center: tuple, diameter: float, angle: float) -> sp.Polygon:
    """
    Create a hexagon centered at the origin, with a specified rotation angle (in degrees)
    and diameter (distance between opposite vertices).

    Parameters:
    angle (float): Rotation angle in degrees.
    diameter (float): Diameter of the hexagon (distance between opposite vertices).

    Returns:
    Polygon: A Shapely Polygon object representing the hexagon.
    """
    radius = diameter / 2
    rotation = math.radians(angle + 90)
    # Generate the points for the hexagon
    points = []
    center_x, center_y = center
    for i in range(6):
        angle_rad = math.radians(60 * i)  # 60 degrees between the points of a hexagon
        x = center_x + radius * math.cos(angle_rad + rotation)
        y = center_y + radius * math.sin(angle_rad + rotation)
        points.append((x, y))

    # Create the hexagon
    hexagon = sp.geometry.Polygon(points)

    return hexagon


def move_point(start: typing.Tuple[float, float], distance: float, direction: float = None, direction_radians: float = None) -> typing.Tuple[float, float]:
    if direction_radians is None:
        direction_radians = math.radians(direction)
    start_x, start_y = start
    delta_x = distance * math.cos(direction_radians)
    delta_y = distance * math.sin(direction_radians)
    return start_x + delta_x, start_y + delta_y


def create_line_string(start: typing.Tuple[float, float], direction: float, distance: float):
    return sp.LineString([start, move_point(start=start,
                                            direction=direction,
                                            distance=distance)])


def polygon_to_linestrings(polygon):
    """
    Break a Polygon into individual LineStrings representing its edges.

    Parameters:
    polygon (Polygon): The input Polygon object.
    include_holes (bool): If True, include the interior boundaries (holes) as well.

    Returns:
    List[LineString]: A list of LineString objects representing the edges of the polygon.
    """
    linestrings = []

    # Process the exterior ring
    exterior_coords = list(polygon.exterior.coords)
    for i in range(1, len(exterior_coords)):
        linestrings.append(sp.LineString([exterior_coords[i - 1], exterior_coords[i]]))
    return linestrings


def polygons_to_sides(polygons):
    """
    Break a Polygon into individual LineStrings representing its edges.

    Parameters:
    polygon (Polygon): The input Polygon object.
    include_holes (bool): If True, include the interior boundaries (holes) as well.

    Returns:
    List[LineString]: A list of LineString objects representing the edges of the polygon.
    """
    vertices = []
    vertices_sides = []

    def add_vertex(vertex):
        i = 0
        for i, v in enumerate(vertices):
            if v.distance(vertex) <= .001:
                break
        else:
            i = len(vertices)
            vertices.append(vertex)
            vertices_sides.append([])
        return i

    sides_vertices = []

    def find_side(sv):
        for i, (a, b) in enumerate(sides_vertices):
            if (a, b) == sv or (b, a) == sv:
                return i
        return -1

    internal_sides = []
    # Process the exterior ring
    for polygon in polygons:
        exterior_coords = list(polygon.exterior.coords)
        origin = add_vertex(sp.Point(exterior_coords[0]))
        vertices_sides.append([])
        point_a = origin
        for i in range(1, len(exterior_coords) - 1):
            point_b = add_vertex(sp.Point(exterior_coords[i]))
            i = find_side((point_a, point_b))
            if i == -1:
                side_number = len(sides_vertices)
                sides_vertices.append((point_a, point_b))
                vertices_sides[point_a].append(side_number)
                vertices_sides[point_b].append(side_number)
            else:
                internal_sides.append(i)
            point_a = point_b
        i = find_side((point_a, origin))
        if i == -1:
            side_number = len(sides_vertices)
            sides_vertices.append((point_a, origin))
            vertices_sides[point_a].append(side_number)
            vertices_sides[origin].append(side_number)
        else:
            internal_sides.append(i)

    filtered_sides_vertices = []
    for i, side_vertices in enumerate(sides_vertices):
        if i not in internal_sides:
            filtered_sides_vertices.append(side_vertices)

    sides = []
    sides_centroids = []
    for a, b in filtered_sides_vertices:
        side = sp.LineString([vertices[a], vertices[b]])
        sides.append(side)
        sides_centroids.append(side.centroid)

    return sides, sides_centroids, vertices, filtered_sides_vertices


from itertools import combinations


def max_distance_between_vertices(polygon):
    """
    Calculate the maximum distance between any two vertices of a polygon's exterior.

    Parameters:
    polygon (Polygon): The input Polygon object.

    Returns:
    float: The maximum distance between any two vertices of the polygon's exterior.
    """
    # Get the exterior coordinates of the polygon
    exterior_coords = list(polygon.exterior.coords)

    # Initialize the maximum distance to be the minimum possible value
    max_distance = 0

    # Iterate over all combinations of two points among the vertices
    for point1, point2 in combinations(exterior_coords, 2):
        # Calculate the distance between the two points
        distance = sp.Point(point1).distance(point2)

        # Update the maximum distance if the current distance is greater
        if distance > max_distance:
            max_distance = distance

    return max_distance


def theta_in_between(theta, start, end):
    if end > start:
        return start < theta < end
    return theta > start or theta < end


def distance(src, dst):
    return math.sqrt((src[0]-dst[0]) ** 2 + (src[1]-dst[1]) ** 2)


def direction(src, dst):
    return math.degrees(math.atan2(dst[1] - src[1], dst[0] - src[0]))


def normalize_direction(direction: float):
    while direction < -180:
        direction += 360
    while direction > 180:
        direction -= 360
    return direction


def direction_difference(direction1: float, direction2: float):
    direction1 = normalize_direction(direction1)
    direction2 = normalize_direction(direction2)
    difference = direction2 - direction1
    if difference > 180:
        difference -= 360
    if difference < -180:
        difference += 360
    return difference


def direction_error_normalization(direction_error: float):
    pi_err = direction_error / 8
    return 1 / (pi_err * pi_err + 1)



import math
import typing


class Navigation:
    def __init__(self,
                 locations: typing.List[typing.Optional[typing.Tuple[float, float]]],
                 paths: typing.List[typing.List[int]],
                 visibility: typing.List[typing.List[typing.List[int]]]):
        self.locations = locations
        self.paths = paths
        self.visibility = visibility

    def closest_location(self,
                         location: typing.Tuple[float, float]) -> int:
        min_dist2 = math.inf
        closest = None
        for i, l in enumerate(self.locations):
            if l is None:
                continue
            dist2 = (l[0] - location[0]) ** 2 + (l[1] - location[1]) ** 2
            if dist2 < min_dist2:
                closest = i
                min_dist2 = dist2
        return closest

    def get_path(self,
                 src: typing.Tuple[float, float],
                 dst: typing.Tuple[float, float]) -> typing.List[typing.Tuple[float, float]]:
        src_index = self.closest_location(location=src)
        dst_index = self.closest_location(location=dst)
        current = src_index
        last_step = src_index
        path_indexes = []
        while current is not None and current != dst_index:
            next_step = self.paths[current][dst_index]
            if next_step == current:
                break
            is_visible = next_step in self.visibility[last_step]
            if not is_visible:
                path_indexes.append(current)
                last_step = current
            current = next_step
        path_indexes.append(dst_index)
        return [self.locations[s] for s in path_indexes]



import typing
import cellworld as cw
from .util import create_hexagon
from .navigation import Navigation


class CellWorldLoader:
    def __init__(self,
                 world_name: str):
        self.world = cw.World.get_from_parameters_names(world_configuration_name="hexagonal",
                                                        world_implementation_name="canonical",
                                                        occlusions_name=world_name)
        paths_builder = cw.Paths_builder.get_from_name(world_configuration_name="hexagonal",
                                                       occlusions_name="21_05")
        paths = cw.Paths(builder=paths_builder, world=self.world)
        cellmap = cw.Cell_map(self.world.configuration.cell_coordinates)

        self.locations: typing.List[typing.Optional[typing.Tuple[float, float]]] = [None
                                                                                    if c.occluded
                                                                                    else tuple(c.location.get_values())
                                                                                    for c in self.world.cells]

        locations_paths: typing.List[typing.List[typing.Optional[int]]] = [[None for _ in range(len(self.world.cells))]
                                                                            for _ in range(len(self.world.cells))]
        for src_cell in self.world.cells:
            for dst_cell in self.world.cells:
                move = paths.get_move(src_cell=src_cell, dst_cell=dst_cell)
                next_step_id = cellmap[src_cell.coordinates + move]
                locations_paths[src_cell.id][dst_cell.id] = next_step_id
        self.paths = locations_paths
        self.open_locations = [tuple(c.location.get_values()) for c in self.world.cells.free_cells()]
        arena_center = self.world.implementation.space.center.get_values()
        arena_transformation: cw.Transformation = self.world.implementation.space.transformation
        cell_transformation: cw.Transformation = self.world.implementation.cell_transformation
        self.arena = create_hexagon(arena_center, arena_transformation.size, arena_transformation.rotation)
        self.occlusions = [create_hexagon(cell.location.get_values(),
                                          cell_transformation.size,
                                          arena_transformation.rotation + cell_transformation.rotation)
                           for cell
                           in self.world.cells.occluded_cells()]
        spawn_cells = cw.Cell_group_builder.get_from_name("hexagonal",
                                                          world_name,
                                                          "spawn_locations")
        self.robot_start_locations = [tuple(self.world.cells[sc].location.get_values()) for sc in spawn_cells]
        self.full_action_list = self.open_locations

        lppo_cells = cw.Cell_group_builder.get_from_name("hexagonal",
                                                         world_name,
                                                         "lppo")

        self.tlppo_action_list = [tuple(self.world.cells[sc].location.get_values()) for sc in lppo_cells]

        cell_visibility = cw.get_resource("graph", "hexagonal", world_name, "cell_visibility")

        self.navigation = Navigation(locations=self.locations,
                                     paths=self.paths,
                                     visibility=cell_visibility)



import math
import typing
import pygame
import shapely as sp
from .resources import Resources
from .util import create_hexagon, move_point, distance
from shapely.affinity import rotate, translate


class CoordinateConverter(object):
    def __init__(self,
                 screen_size: typing.Tuple[int, int],
                 flip_y: bool = False):
        self.screen_size = screen_size
        self.screen_width, self.screen_height = screen_size
        self.flip_y = flip_y
        self.screen_offset = (self.screen_width - self.screen_height) / 2
        self.hexa_ratio = (math.sqrt(3) / 2)

    def from_canonical(self, canonical: tuple):
        canonical_x, canonical_y = canonical
        screen_x = canonical_x * self.screen_width
        if self.flip_y:
            screen_y = (1-canonical_y) * self.screen_width - self.screen_offset
        else:
            screen_y = canonical_y * self.screen_width - self.screen_offset
        return screen_x, screen_y

    def to_canonical(self, screen_x: int, screen_y: int):
        y = self.screen_height - screen_y + self.screen_offset
        canonical_y = y / self.screen_height * self.hexa_ratio
        canonical_x = screen_x / self.screen_width
        return canonical_x, canonical_y


class AgentState(object):
    def __init__(self, location: typing.Tuple[float, float] = (0, 0), direction: float = 0):
        self.location = location
        self.direction = direction

    def __iter__(self):
        yield self.location
        yield self.direction

    def update(self,
               distance: float,
               rotation: float) -> "AgentState":
        new_direction = self.direction + rotation
        return AgentState(location=move_point(start=self.location,
                                              direction=new_direction,
                                              distance=distance),
                          direction=new_direction)


class AgentDynamics(object):
    def __init__(self, forward_speed: float, turn_speed: float):
        self.forward_speed = forward_speed
        self.turn_speed = turn_speed

    def __iter__(self):
        yield self.forward_speed
        yield self.turn_speed

    def change(self, delta_t: float) -> tuple:
        return self.forward_speed * delta_t,  self.turn_speed * delta_t


class Agent(object):

    def __init__(self,
                 view_field: float = 180,
                 collision: bool = True):
        self.view_field = view_field
        self._state: AgentState = AgentState()
        self.dynamics: AgentDynamics = AgentDynamics(forward_speed=0,
                                                     turn_speed=0)
        self.polygon = self.create_polygon()
        self.sprite = None
        self.collision = collision
        self.on_reset = None
        self.on_step = None
        self.on_start = None
        self.name = ""
        self.model = None
        self.trajectory: typing.List[AgentState] = []
        self.coordinate_converter: typing.Optional[CoordinateConverter] = None
        self.running = False

    def set_sprite_size(self, size: tuple):
        self.sprite = pygame.transform.scale(self.create_sprite(), size)

    def set_state(self, state: AgentState) -> None:
        self.trajectory.append(state)
        self._state = state

    @property
    def state(self) -> AgentState:
        return self._state

    def reset(self) -> None:
        self.trajectory.clear()
        if self.on_reset:
            self.on_reset()
        self.running = True

    def start(self) -> None:
        if self.on_start:
            self.on_start()

    def step(self, delta_t: float) -> None:
        if self.on_step:
            self.on_step(delta_t)

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("agent.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 90)
        return rotated_sprite

    @staticmethod
    def create_polygon() -> sp.Polygon:
        return create_hexagon((0, 0), .05, 30)

    def get_polygon(self,
                    state: AgentState = None) -> sp.Polygon:
        # Rotate and then translate the arrow polygon
        if state:
            x, y = state.location
            direction = state.direction
        else:
            x, y = self._state.location
            direction = self._state.direction
        rotated_polygon = rotate(self.polygon,
                                 direction,
                                 origin=(0, 0),
                                 use_radians=False)
        translated_polygon = translate(rotated_polygon, x, y)
        return translated_polygon

    def get_sprite(self) -> pygame.Surface:
        rotated_sprite = pygame.transform.rotate(self.sprite, self._state.direction)
        return rotated_sprite

    def get_observation(self) -> dict:
        if self.model:
            return self.model.get_observation(agent_name=self.name)
        else:
            return None

    def get_stats(self) -> dict:
        stats = {}
        dist = 0
        prev_state = self.trajectory[0]
        for state in self.trajectory[1:]:
            dist += distance(prev_state.location, state.location)
            prev_state = state
        stats["distance"] = dist
        return stats

    def render(self,
               surface: pygame.Surface,
               coordinate_converter: CoordinateConverter):
        agent_sprite: pygame.Surface = self.get_sprite()
        width, height = agent_sprite.get_size()
        screen_x, screen_y = coordinate_converter.from_canonical(self.state.location)
        surface.blit(agent_sprite, (screen_x - width / 2, screen_y - height / 2))

    def set_coordinate_converter(self,
                                 coordinate_converter: CoordinateConverter):
        self.coordinate_converter = coordinate_converter



import math
import time
import typing
import shapely as sp
from .agent import Agent, AgentState
from .visibility import Visibility
from .geometry import atan2


class Model(object):

    def __init__(self,
                 arena,
                 occlusions,
                 real_time: bool = False,
                 time_step: float = 0.1):
        self.arena = arena
        self.occlusions = occlusions
        self.real_time = real_time
        self.time_step = time_step
        self.agents: typing.Dict[str, Agent] = {}
        self.visibility = Visibility(arena=self.arena, occlusions=self.occlusions)
        self.last_step = None
        self.time = 0
        self.running = False
        self.step_count = 0

    def set_agents_state(self, agents_state: typing.Dict[str, AgentState],
                         delta_t: float = 0):
        for agent_name, agent_state in agents_state.items():
            if agent_name in self.agents:
                self.agents[agent_name].set_state(agent_state)
        self.__update_state__(delta_t)

    def __update_state__(self,
                         delta_t: float = 0):
        pass

    def add_agent(self, name: str, agent: Agent):
        self.agents[name] = agent
        agent.name = name
        agent.model = self

    def reset(self):
        self.running = True
        for name, agent in self.agents.items():
            agent.reset()
        observations = self.get_observations()
        for name, agent in self.agents.items():
            agent.start()
        self.last_step = time.time()
        self.step_count = 0

    def is_valid_state(self, agent_polygon: sp.Polygon, collisions: bool) -> bool:
        if not self.arena.contains(agent_polygon):
            return False
        if collisions:
            for occlusion in self.occlusions:
                if agent_polygon.intersects(occlusion):
                    return False
        return True

    def wall_direction(self, src: typing.Tuple[float, float], wall_number: int):
        return math.degrees(atan2(sp.Point(src), self.visibility.walls_centroids[wall_number]))

    def get_observations(self):
        agent_visibility = {src_name: {dst_name: None for dst_name in self.agents} for src_name in self.agents}
        observations = {}
        for src_name, src_agent in self.agents.items():
            observations[src_name] = {}
            src_point = sp.Point(src_agent.state.location)
            walls_by_distance = self.visibility.walls_by_distance(src=src_point)
            parsed_walls = []
            for wall_number, vertices, distance in walls_by_distance:
                parsed_walls.append((distance, self.wall_direction(src=src_point, wall_number=wall_number)))
            observations[src_name]["walls"] = parsed_walls
            for dst_name, dst_agent in self.agents.items():
                if agent_visibility[src_name][dst_name] is None:
                    if src_name == dst_name:
                        is_visible = True
                    else:
                        dst_point = sp.Point(dst_agent.state.location)
                        is_visible = self.visibility.line_of_sight(src=src_point,
                                                                   dst=dst_point,
                                                                   walls_by_distance=walls_by_distance)
                    agent_visibility[src_name][dst_name] = is_visible
                    agent_visibility[dst_name][src_name] = is_visible
            observations[src_name]["agent_states"] = {}
            for dst_name, is_visible in agent_visibility.items():
                if is_visible:
                    observations[src_name]["agent_states"][dst_name] = self.agents[dst_name].state.location, self.agents[dst_name].state.direction
                else:
                    observations[src_name]["agent_states"][dst_name] = None
        return observations

    def get_observation(self,
                        agent_name: str,
                        polygonal: bool = False) -> dict:
        observation = {}
        src_point = sp.Point(self.agents[agent_name].state.location)
        if polygonal:
            visibility_polygon = self.visibility.get_visibility_polygon(src_point.state.location, src_point.state.direction)
        else:
            walls_by_distance = self.visibility.walls_by_distance(src=src_point)
        parsed_walls = []
        for wall_number, vertices, distance in walls_by_distance:
            parsed_walls.append((distance, self.wall_direction(src=src_point, wall_number=wall_number)))
        observation["walls"] = parsed_walls
        observation["agent_states"] = {}
        for dst_name, dst_agent in self.agents.items():
            if agent_name == dst_name:
                is_visible = True
            else:
                dst_point = sp.Point(dst_agent.state.location)
                if polygonal:
                    dst_polygon = dst_agent.get_polygon()
                    is_visible = dst_polygon.intersects(visibility_polygon)
                else:
                    is_visible = self.visibility.line_of_sight(src=src_point,
                                                               dst=dst_point,
                                                               walls_by_distance=walls_by_distance)
            if is_visible:
                observation["agent_states"][dst_name] = self.agents[dst_name].state.location, self.agents[dst_name].state.direction
            else:
                observation["agent_states"][dst_name] = None
        return observation

    def step(self) -> float:
        if self.real_time:
            while self.last_step + self.time_step > time.time():
                pass

        self.last_step = time.time()
        for name, agent in self.agents.items():
            dynamics = agent.dynamics
            distance, rotation = dynamics.change(delta_t=self.time_step)
            new_state = agent.state.update(rotation=rotation,
                                           distance=distance)
            agent_polygon = agent.get_polygon(state=new_state)
            if self.is_valid_state(agent_polygon=agent_polygon,
                                   collisions=agent.collision):
                agent.set_state(state=new_state)
            else: #try only rotation
                new_state = agent.state.update(rotation=rotation,
                                               distance=0)
                agent_polygon = agent.get_polygon(state=new_state)
                if self.is_valid_state(agent_polygon=agent_polygon,
                                       collisions=agent.collision):
                    agent.set_state(state=new_state)
                else: #try only translation
                    new_state = agent.state.update(rotation=0,
                                                   distance=distance)
                    agent_polygon = agent.get_polygon(state=new_state)
                    if self.is_valid_state(agent_polygon=agent_polygon,
                                           collisions=agent.collision):
                        agent.set_state(state=new_state)
        for name, agent in self.agents.items():
            agent.step(delta_t=self.time_step)
        self.time += self.time_step
        self.step_count += 1
        return self.time_step



import random
from .util import distance
from .model import Model
from .agent import AgentState, CoordinateConverter
from .mouse import Mouse
from .robot import Robot
from .cellworld_loader import CellWorldLoader


class BotEvade(Model):
    def __init__(self,
                 world_name: str = "21_05",
                 use_predator: bool = True,
                 puff_cool_down_time: float = .5,
                 puff_threshold: float = .1,
                 goal_location=(1.0, 0.5),
                 goal_threshold: float = .1,
                 time_step: float = .025,
                 real_time: bool = False,
                 render: bool = False):
        self.use_predator = use_predator
        self.puff_cool_down_time = puff_cool_down_time
        self.puff_threshold = puff_threshold
        self.goal_location = goal_location
        self.goal_threshold = goal_threshold
        self.render = render
        self.loader = CellWorldLoader(world_name=world_name)

        Model.__init__(self,
                       arena=self.loader.arena,
                       occlusions=self.loader.occlusions,
                       time_step=time_step,
                       real_time=real_time)

        if use_predator:
            self.predator = Robot(start_locations=self.loader.robot_start_locations,
                                  open_locations=self.loader.open_locations,
                                  navigation=self.loader.navigation)

            self.add_agent("predator", self.predator)

        self.prey = Mouse(start_state=AgentState(location=(.05, .5),
                                                 direction=0),
                          navigation=self.loader.navigation)

        self.add_agent("prey", self.prey)

        self.running = False

        if self.render:
            from .view import View
            self.view = View(model=self)
            self.view.on_quit = self.__on_quit__

            if use_predator:
                import pygame

                def render_puff_area(surface: pygame.Surface,
                                     coordinate_converter: CoordinateConverter):
                    predator_location = coordinate_converter.from_canonical(self.predator.state.location)
                    puff_area_size = self.puff_threshold * coordinate_converter.screen_width
                    puff_location = predator_location[0] - puff_area_size, predator_location[1] - puff_area_size
                    puff_area_surface = pygame.Surface((puff_area_size * 2, puff_area_size * 2), pygame.SRCALPHA)
                    puff_area_color = (255, 0, 0, 60) if self.puff_cool_down > 0 else (0, 0, 255, 60)
                    pygame.draw.circle(puff_area_surface,
                                       color=puff_area_color,
                                       center=(puff_area_size, puff_area_size),
                                       radius=puff_area_size)
                    surface.blit(puff_area_surface,
                                 puff_location)
                    pygame.draw.circle(surface=surface,
                                       color=(0, 0, 255),
                                       center=predator_location,
                                       radius=puff_area_size,
                                       width=2)

                self.view.add_render_step(render_puff_area)

        self.puffed: bool = False
        self.puff_cool_down: float = 0
        self.goal_achieved: bool = True
        self.predator_prey_distance: float = 1
        self.prey_goal_distance: float = 0
        self.puff_count = 0

    def __update_state__(self,
                         delta_t: float = 0):
        if self.use_predator and self.puff_cool_down <= 0:
            self.predator_prey_distance = distance(self.prey.state.location,
                                                   self.predator.state.location)
            if self.visibility.line_of_sight(self.prey.state.location, self.predator.state.location):
                if self.predator_prey_distance <= self.puff_threshold:
                    self.puffed = True
                    self.puff_count += 1
                    self.puff_cool_down = self.puff_cool_down_time

                self.predator.set_destination(self.prey.state.location)

            if not self.predator.path:
                self.predator.set_destination(random.choice(self.loader.open_locations))

        if delta_t < self.puff_cool_down:
            self.puff_cool_down -= delta_t
        else:
            self.puff_cool_down = 0

        self.prey_goal_distance = distance(self.goal_location, self.prey.state.location)

        if self.prey_goal_distance <= self.goal_threshold:
            self.goal_achieved = True
            self.running = False

    def __on_quit__(self):
        self.running = False

    def reset(self):
        Model.reset(self)
        self.puff_count = 0
        self.__update_state__()

    def step(self) -> float:
        delta_t = Model.step(self)
        if self.render:
            self.view.draw()
        self.__update_state__(delta_t=delta_t)
        return delta_t


from .view import View
from .model import Model
from .agent import Agent, AgentDynamics, AgentState
from .resources import Resources
from .robot import Robot
from .mouse import Mouse
from .ray_tracing import RayTracing
from .navigation import Navigation
from .util import create_hexagon, distance, direction
from .cellworld_loader import CellWorldLoader
from .visibility import Visibility
from .botevade import BotEvade



import pygame
import math
import typing
import shapely as sp
from .model import Model
from .agent import AgentState
from .util import create_line_string


class RayTracing:
    def __init__(self, model: Model,
                 horizontal_view_field: float = 270.0,
                 vertical_view_field: float = 90,
                 perspective_height: float = .025,
                 resolution: int = 1):
        self.model = model
        self.vertical_view_field = vertical_view_field
        self.horizontal_view_field = horizontal_view_field
        self.perspective_height = perspective_height
        self.resolution = resolution
        self.walls = sp.geometry.LineString(model.arena.exterior)
        self.occlusions = [sp.geometry.LineString(occlusion.exterior) for occlusion in model.occlusions]
        self.ray_length = 1.0
        self.occlusion_color: typing.Tuple[int, int, int] = (255, 255, 255)
        self.occlusion_height: float = 0.1
        self.wall_color: typing.Tuple[int, int, int] = (255, 200, 200)
        self.wall_height: float = 0.5

    def distance_factor(self, distance: float) -> float:
        return (self.ray_length - distance) / self.ray_length

    def line_limits(self, perspective_height, line_bottom, line_top, distance, view_field_offset) -> typing.Tuple[float, float]:
        bottom_angle = self.vertical_view_field / 2 + math.degrees(math.atan2(perspective_height, distance))
        top_angle = self.vertical_view_field / 2 - math.degrees(math.atan2(line_top-perspective_height, distance))
        return bottom_angle / self.vertical_view_field, top_angle / self.vertical_view_field

    def render_line(self, screen, screen_x, distance, view_field_offset: float, color, height):
        screen_width, screen_height = screen.get_size()
        distance_factor = self.distance_factor(distance=distance)
        line_color = tuple(channel * distance_factor for channel in color)
        line_start, line_end = self.line_limits(self.perspective_height,
                                                0,
                                                height,
                                                distance,
                                                view_field_offset)
        y_start = screen_height * line_start
        y_ends = screen_height * line_end
        pygame.draw.line(surface=screen,
                         color=line_color,
                         start_pos=(screen_x, y_start),
                         end_pos=(screen_x, y_ends),
                         width=self.resolution)

    def render(self, perspective: AgentState, screen: pygame.Surface):
        screen_width, screen_height = screen.get_size()
        start_angle = perspective.direction - self.horizontal_view_field / 2
        step_angle = self.horizontal_view_field / screen_width
        start = sp.Point(perspective.location)
        screen.fill((0, 0, 0))
        sorted_occlusions = sorted(self.occlusions, key=lambda line: line.distance(start))
        for x in range(0, screen_width, self.resolution):
            view_angle = step_angle * x
            view_field_offset = view_angle - self.horizontal_view_field / 2
            ray_direction = start_angle + view_angle
            screen_x = screen_width - x - 1
            ray = create_line_string(start=perspective.location,
                                     direction=ray_direction,
                                     distance=self.ray_length)
            wall_intersection = ray.intersection(self.walls)
            wall_distance = wall_intersection.distance(start)
            self.render_line(screen=screen,
                             screen_x=screen_x,
                             color=self.wall_color,
                             height=self.wall_height,
                             distance=wall_distance,
                             view_field_offset=view_field_offset)
            closest_distance = wall_distance
            occluded = False
            for occlusion in sorted_occlusions:
                intersection = ray.intersection(occlusion)
                if intersection:
                    for intersection_point in intersection.geoms:
                        distance = intersection_point.distance(start)
                        if distance < closest_distance:
                            closest_distance = distance
                            occluded = True
                    if occluded:
                        break
            if occluded:
                self.render_line(screen=screen,
                                 screen_x=screen_x,
                                 color=self.occlusion_color,
                                 height=self.occlusion_height,
                                 distance=closest_distance,
                                 view_field_offset=view_field_offset)

        pygame.display.flip()



from setuptools import setup

setup(name='cellworld_game',
      description='Simulation environment for the cellworld research platform',
      url='https://github.com/germanespinosa/cellworld_game',
      author='German Espinosa',
      author_email='germanespinosa@gmail.com',
      long_description=open('./cellworld_game/README.md').read() + '\n---\n<small>Package created with Easy-pack</small>\n',
      long_description_content_type='text/markdown',
      packages=['cellworld_game'],
      install_requires=['cellworld', 'pygame', 'shapely'],
      license='MIT',
      include_package_data=True,
      version='0.0.49',
      zip_safe=False)



import math
import pygame
import typing
from .model import Model
from .agent import CoordinateConverter
from .util import generate_distinct_colors


class View(object):
    def __init__(self,
                 model: Model,
                 screen_width: int = 800,
                 flip_y: bool = True):
        pygame.init()
        self.model = model
        self.visibility = model.visibility
        self.screen_width = screen_width
        self.hexa_ratio = (math.sqrt(3) / 2)
        self.coordinate_converter = CoordinateConverter(screen_size=(screen_width, int(self.hexa_ratio * screen_width)),
                                                        flip_y=flip_y)
        for agent_name, agent in model.agents.items():
            agent.set_sprite_size((screen_width/10, screen_width/10))
            agent.set_coordinate_converter(self.coordinate_converter)
        self.screen = pygame.display.set_mode(self.coordinate_converter.screen_size)
        self.arena_color = (255, 255, 255)
        self.occlusion_color = (50, 50, 50)
        self.agent_colors = {agent_name: color for agent_name, color
                             in zip(self.model.agents.keys(),
                                    generate_distinct_colors(len(self.model.agents)))}
        self.clock = pygame.time.Clock()
        self.agent_perspective = -1
        self.show_sprites = True
        self.target = None
        self.on_mouse_button_down = None
        self.on_mouse_button_up = None
        self.on_mouse_move = None
        self.on_mouse_wheel = None
        self.on_key_up = None
        self.on_quit = None
        self.pressed_keys = pygame.key.get_pressed()
        self.render_steps: typing.List[typing.Callable[[pygame.Surface, CoordinateConverter], None]] = []

    def add_render_step(self,
                        render_step: typing.Callable[[pygame.Surface, CoordinateConverter], None]):
        self.render_steps.append(render_step)

    def draw_polygon(self, polygon, color):
        """Draws a hexagon at the specified position and size."""
        pygame.draw.polygon(self.screen,
                            color,
                            [self.coordinate_converter.from_canonical(point) for point in polygon.exterior.coords])

    def draw_polygon_vertices(self, polygon, color, size=2):
        """Draws a hexagon at the specified position and size."""
        for point in polygon.exterior.coords:
            pygame.draw.circle(surface=self.screen,
                               color=color,
                               center=self.coordinate_converter.from_canonical(point),
                               radius=size,
                               width=2)

    def draw_points(self, points, color, size=2):
        """Draws a hexagon at the specified position and size."""
        for point in points:
            pygame.draw.circle(surface=self.screen,
                               color=color,
                               center=self.coordinate_converter.from_canonical((point.x, point.y)),
                               radius=size,
                               width=2)

    def draw(self):
        self.screen.fill((0, 0, 0))
        self.draw_polygon(self.model.arena, self.arena_color)

        if self.agent_perspective != -1:
            agent_name = list(self.model.agents.keys())[self.agent_perspective]
            visibility_perspective = self.model.agents[agent_name].state
            visibility_polygon, a = self.visibility.get_visibility_polygon(location=visibility_perspective.location,
                                                                           direction=visibility_perspective.direction,
                                                                           view_field=360)
            self.draw_polygon(visibility_polygon, (180, 180, 180))

        for occlusion in self.model.occlusions:
            self.draw_polygon(occlusion, self.occlusion_color)

        for (name, agent), color in zip(self.model.agents.items(), self.agent_colors):
            if self.show_sprites:
                agent.render(surface=self.screen,
                             coordinate_converter=self.coordinate_converter)
            else:
                self.draw_polygon(self.model.agents[name].get_polygon(), color=self.agent_colors[name])

        for render_step in self.render_steps:
            render_step(self.screen, self.coordinate_converter)

        self.__process_events__()
        pygame.display.flip()

    def __process_events__(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                if self.on_quit:
                    self.on_quit()
            if event.type == pygame.MOUSEBUTTONDOWN:
                x, y = event.pos
                canonical_x_y = self.coordinate_converter.to_canonical(x, y)
                if self.on_mouse_button_down:
                    self.on_mouse_button_down(event.button, canonical_x_y)
            elif event.type == pygame.MOUSEBUTTONUP:
                x, y = event.pos
                canonical_x_y = self.coordinate_converter.to_canonical(x, y)
                if self.on_mouse_button_up:
                    self.on_mouse_button_up(event.button, canonical_x_y)
            elif event.type == pygame.MOUSEMOTION:
                x, y = event.pos
                canonical_x_y = self.coordinate_converter.to_canonical(x, y)
                if self.on_mouse_move:
                    self.on_mouse_move(canonical_x_y)
            elif event.type == pygame.MOUSEWHEEL:
                x, y = event.pos
                canonical_x_y = self.coordinate_converter.to_canonical(x, y)
                if self.on_mouse_wheel:
                    self.on_mouse_wheel(event.button, canonical_x_y)
            elif event.type == pygame.KEYDOWN:
                if self.on_key_down:
                    self.on_key_down(key=event.key)
            elif event.type == pygame.KEYUP:
                if self.on_key_up:
                    self.on_key_up(event.key)
        self.pressed_keys = pygame.key.get_pressed()

    def on_key_down(self, key):
        if key == pygame.K_0:
            self.agent_perspective = -1
        if key == pygame.K_1:
            self.agent_perspective = 0
        if key == pygame.K_2:
            self.agent_perspective = 1
        if key == pygame.K_3:
            self.show_sprites = False
        if key == pygame.K_4:
            self.show_sprites = True



import pygame
from .util import distance, direction, direction_difference, direction_error_normalization
from .agent import Agent, CoordinateConverter
from .navigation import Navigation


class NavigationAgent(Agent):

    def __init__(self,
                 navigation: Navigation,
                 max_forward_speed: float,
                 max_turning_speed: float,
                 threshold: float = 0.01,
                 destination_update_rate: int = 10):
        self.max_forward_speed = max_forward_speed
        self.max_turning_speed = max_turning_speed
        self.threshold = threshold
        self.destination_update_rate = destination_update_rate
        self.navigation = navigation
        self.new_destination = None
        self.destination = None
        self.navigation_plan_update_wait = 0
        self.destination_wait = 0
        self.path = []
        Agent.__init__(self)
        self.collision = False

    def next_step(self):
        if self.path:
            return self.path[0]
        return None

    def set_destination(self, destination):
        self.new_destination = destination

    def reset(self):
        self.destination = None
        self.path = []
        Agent.reset(self)

    def start(self):
        Agent.start(self)

    def step(self, delta_t: float):
        if not self.running:
            self.stop_navigation()
            return

        if self.new_destination and self.destination_wait == 0:
            self.destination = self.new_destination
            self.new_destination = None
            self.destination_wait = self.destination_update_rate
            self.navigation_plan_update_wait = self.destination_update_rate
            self.path = self.navigation.get_path(src=self.state.location, dst=self.destination)

        if self.destination and self.navigation_plan_update_wait == 0:
            self.path = self.navigation.get_path(src=self.state.location, dst=self.destination)

        if self.destination_wait:
            self.destination_wait -= 1

        if self.navigation_plan_update_wait:
            self.navigation_plan_update_wait -= 1

        if self.next_step() is not None:
            distance_error = distance(src=self.state.location,
                                      dst=self.next_step())
            if distance_error < self.threshold:
                self.path.pop(0)

        if self.next_step():
            distance_error = distance(src=self.state.location,
                                      dst=self.next_step())

            normalized_distance_error = max(distance_error/.2, 1)

            destination_direction = direction(src=self.state.location,
                                              dst=self.next_step())
            direction_error = direction_difference(direction1=self.state.direction,
                                                   direction2=destination_direction)
            normalized_direction_error = direction_error_normalization(direction_error=direction_error)

            self.dynamics.forward_speed = self.max_forward_speed * normalized_direction_error * normalized_distance_error
            self.dynamics.turn_speed = self.max_turning_speed * direction_error
        else:
            self.dynamics.forward_speed = 0
            self.dynamics.turn_speed = 0

    def render(self,
               surface: pygame.Surface,
               coordinate_converter: CoordinateConverter):
        current_point = self.state.location
        for step in self.path:
            pygame.draw.line(surface,
                             (255, 0, 0),
                             coordinate_converter.from_canonical(current_point),
                             coordinate_converter.from_canonical(step),
                             2)
            pygame.draw.circle(surface=surface,
                               color=(0, 0, 255),
                               center=coordinate_converter.from_canonical(step),
                               radius=5,
                               width=2)
            current_point = step
        Agent.render(self=self,
                     surface=surface,
                     coordinate_converter=coordinate_converter)

    def stop_navigation(self):
        self.dynamics.forward_speed = 0
        self.dynamics.turn_speed = 0
